<input type="checkbox" class="{{ default_classes }} accent-cyan-600 shadow-slate-50 h-4 w-4"
              {% if default == true or (default and default.lower() in ["true", "t"]) %}checked{% endif %}
              {{ default_input_props|safe }}
              {{ readonly_attr|safe }}
>